(*  Title: 	HOL/subset
    Author: 	Lawrence C Paulson, Cambridge University Computer Laboratory
    Copyright   1991  University of Cambridge

Derived rules involving subsets
Union and Intersection as lattice operations
*)

(*** Big Union -- least upper bound of a class  ***)

val prems = goal Class.thy
    "B:A ==> B <= Union(A)";
by (REPEAT (ares_tac (prems@[subsetI,UnionI]) 1));
val Union_upper = result();

val prems = goal Class.thy
    "[| X:A;  C<=X |] ==> C <= Union(A)";
by (rtac subset_trans 1);
by (REPEAT (resolve_tac (prems@[Union_upper]) 1));
val Union_upperbound = result();

val prems = goal Class.thy
    "[| !!X. X:A ==> X<=C |] ==> Union(A) <= C";
by (REPEAT (ares_tac [subsetI] 1
     ORELSE eresolve_tac ([UnionE] @ (prems RL [subsetD])) 1));
val Union_least = result();


(*** Big Intersection -- greatest lower bound of a class ***)

val prems = goal Class.thy
    "B:A ==> Inter(A) <= B";
by (REPEAT (resolve_tac (prems@[subsetI]) 1
     ORELSE etac InterD 1));
val Inter_lower = result();

val prems = goal Class.thy
    "[| X:A;  X<=C |] ==> Inter(A) <= C";
by (rtac subset_trans 1);
by (REPEAT (resolve_tac (prems@[Inter_lower]) 1));
val Inter_lowerbound = result();

val prems = goal Class.thy
    "[| !!X. X:A ==> C<=X |] ==> C <= Inter(A)";
by (REPEAT (ares_tac [subsetI,InterI] 1
     ORELSE eresolve_tac (prems RL [subsetD]) 1));
val Inter_greatest = result();

(*** Finite Union -- the least upper bound of 2 classes ***)

goal Class.thy "A <= A Un B";
by (REPEAT (ares_tac [subsetI,UnI1] 1));
val Un_upper1 = result();

goal Class.thy "B <= A Un B";
by (REPEAT (ares_tac [subsetI,UnI2] 1));
val Un_upper2 = result();

val prems = goal Class.thy "C<=A ==> C <= A Un B";
by (rtac subset_trans 1);
by (resolve_tac prems 1);
by (rtac Un_upper1 1);
val Un_upperbound1 = result();

val prems = goal Class.thy "C<=B ==> C <= A Un B";
by (rtac subset_trans 1);
by (resolve_tac prems 1);
by (rtac Un_upper2 1);
val Un_upperbound2 = result();

val prems = goal Class.thy "[| A<=C;  B<=C |] ==> A Un B <= C";
by (cut_facts_tac prems 1);
by (DEPTH_SOLVE (ares_tac [subsetI] 1 
          ORELSE eresolve_tac [UnE,subsetD] 1));
val Un_least = result();

(*** Finite Intersection -- the greatest lower bound of 2 classes *)

goal Class.thy "A Int B <= A";
by (REPEAT (ares_tac [subsetI] 1 ORELSE etac IntE 1));
val Int_lower1 = result();

goal Class.thy "A Int B <= B";
by (REPEAT (ares_tac [subsetI] 1 ORELSE etac IntE 1));
val Int_lower2 = result();

val prems = goal Class.thy "A<=C ==> A Int B <= C";
by (resolve_tac [Int_lower1 RS subset_trans] 1);
by (resolve_tac prems 1);
val Int_lowerbound1 = result();

val prems = goal Class.thy "B<=C ==> A Int B <= C";
by (resolve_tac [Int_lower2 RS subset_trans] 1);
by (resolve_tac prems 1);
val Int_lowerbound2 = result();

val prems = goal Class.thy "[| C<=A;  C<=B |] ==> C <= A Int B";
by (cut_facts_tac prems 1);
by (REPEAT (ares_tac [subsetI,IntI] 1
     ORELSE etac subsetD 1));
val Int_greatest = result();

val class_cs = HOL_cs 
    addSIs [ballI, subsetI, InterI, INT_I, CollectI, ComplI, IntI, UnCI] 
    addIs  [bexI, UnionI, UN_I, singletonI, imageI, rangeI] 
    addXIs [bexCI] 
    addSEs [bexE, UnionE, UN_E,
	    CollectE, NComplE, IntE, UnE, singletonE, imageE, rangeE ] 
    addEs  [ballE, InterD, InterE, INT_D, INT_E, subsetD, subsetCE];

fun cfast_tac prems = cut_facts_tac prems THEN' fast_tac class_cs;
