(*  Title:      ZF/ex/simpdata
    Author:     Lawrence C Paulson, Cambridge University Computer Laboratory
    Copyright   1991  University of Cambridge

Rewriting for ZF set theory -- based on FOL rewriting
*)

fun prove_fun s = 
    (writeln s;  prove_goal set_thy s
       (fn prems => [ (cut_facts_tac prems 1),
		      (fast_tac ZF_cs 1) ]));

val anti_refl_rew = prove_goal set_thy "a:a <-> False"
 (K [ (fast_tac (ZF_cs addEs [mem_anti_refl]) 1) ]);

val mem_rews = map prove_fun
 [ "a:0 <-> False",
   "a : A Un B <-> a:A | a:B",
   "a : A Int B <-> a:A & a:B",
   "a : A-B <-> a:A & ~a:B",
   "a : cons(b,B) <-> a=b | a:B",
   "i : succ(j) <-> i=j | i:j",
   "A: Union(C) <-> (EX B. B:C & A:B)",
   "(EX B. B:C) ==> A: Inter(C) <-> (ALL B. B:C --> A:B)",
   "a : Collect(A,P) <-> a:A & P(a)",
   "b : RepFun(A,f) <-> (EX x. x:A & b=f(x))" ];

(** Tactics for type checking -- from CTT **)

fun is_rigid_elem (Const("Trueprop",_) $ (Const("op :",_) $ a $ _)) = 
      not (is_var (head_of a))
  | is_rigid_elem _ = false;

(*Try solving a:A by assumption provided a is rigid!*) 
val test_assume_tac = SUBGOAL(fn (prem,i) =>
    if is_rigid_elem (Logic.strip_assums_concl prem)
    then  assume_tac i  else  no_tac);

(*Type checking: solve a:A (a rigid, A flexible). *)
fun typechk_tac thms =
  REPEAT_FIRST (test_assume_tac ORELSE' 
		compat_resolve_tac thms 3);

val ZF_typechecks = [lam_type,SigmaI,apply_type,split_type];

fun auto_step_tac tyrls = test_assume_tac ORELSE' compat_resolve_tac tyrls 3;
fun auto_step_tac tyrls = ares_tac tyrls;
fun auto_tac tyrls = SELECT_GOAL (DEPTH_SOLVE_1 (auto_step_tac tyrls 1));

(** New version of mk_rew_rules -- at least, must handle bounded quants! **)

(*Could check whether subsetD, CollectD1, CollectD2, singleton_subsetD
  are applicable... *)
fun ZF_atomize th = case concl_of th of
	_ $ (Const("Ball",_) $ _ $ _) => atomize(th RS bspec)
    |	_ => atomize th;

fun ZF_mk_rew_rules th = map mk_eq (ZF_atomize th);


val standard_auto_tac = auto_tac [TrueI,ballI];
val auto_tac_ref = ref(standard_auto_tac);

structure ZF_SimpData : SIMP_DATA =
  struct
  val refl_thms		= FOL_SimpData.refl_thms
  val trans_thms	= FOL_SimpData.trans_thms
  val red1		= FOL_SimpData.red1
  val red2		= FOL_SimpData.red2
  val mk_rew_rules	= ZF_mk_rew_rules 
  fun auto_tac i 	= !auto_tac_ref i  (*delayed dereferencing!!!*)
  val case_splits	= FOL_SimpData.case_splits
  val norm_thms		= FOL_SimpData.norm_thms
  val subst_thms	= FOL_SimpData.subst_thms
  val dest_red		= FOL_SimpData.dest_red
  end;

structure ZF_Simp = SimpFun(ZF_SimpData);

open ZF_Simp;

(*Redeclared because the previous FOL_ss belongs to a different instance
  of type simpset*)
val FOL_ss = empty_ss addcongs FOL_congs addrews FOL_rews;

(** Basic congruence and rewrite rules for ZF set theory **)

val ZF_congs = 
   [ball_cong,bex_cong,Replace_cong,RepFun_cong,Collect_cong,the_cong,
    Sigma_cong,split_cong,Pi_cong,lam_cong] @ basic_ZF_congs;

val ZF_rews = [ball_rew, beta_conv, eta_conv, restrict_conv,
	       fst_conv, snd_conv, split_conv];

val ZF_ss = FOL_ss addcongs ZF_congs
                   addrews (ZF_rews@[anti_refl_rew]@mem_rews);

(*simplify all subgoals WRT extra rewrite rules*)
fun all_simp_tac ss rews = ALLGOALS (ASM_SIMP_TAC (ss addrews rews));
